#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

// Fused kernel that computes per-feature reduction to obtain mean and variance
// and then applies batch normalization, bias addition, division and Swish activation.
// Each block is assigned to one feature (column) of the x_linear tensor of shape [batch_size, out_features].
// This design avoids unnecessary atomic operations by confining reductions and running stat updates to a single block per feature.

// Template kernel
template <typename scalar_t>
__global__ void fused_bn_swish_kernel(
    const scalar_t* __restrict__ x_linear, // Input: result of linear transformation, shape [batch_size, out_features]
    scalar_t* __restrict__ output,         // Output tensor, same shape as x_linear
    const scalar_t* __restrict__ bn_weight,  // BatchNorm weight, shape [out_features]
    const scalar_t* __restrict__ bn_bias,    // BatchNorm bias, shape [out_features]
    scalar_t* __restrict__ bn_running_mean,  // Running mean, shape [out_features]
    scalar_t* __restrict__ bn_running_var,   // Running variance, shape [out_features]
    const scalar_t* __restrict__ add_bias,   // Additional bias (1-element tensor)
    const float bn_eps,                      // BatchNorm epsilon
    const float bn_momentum,                 // BatchNorm momentum
    const float divide_value,                // Division value
    const int batch_size,
    const int out_features) {

    // Each block handles one feature column
    int f = blockIdx.x;
    if (f >= out_features) return;
    int tid = threadIdx.x;

    // Allocate shared memory for reduction: first half for sum, second half for sum of squares
    extern __shared__ float shared_data[]; // size: 2 * blockDim.x
    float* s_sum = shared_data;              // shared array for sums
    float* s_sumsq = shared_data + blockDim.x; // shared array for sum of squares

    float local_sum = 0.0f;
    float local_sumsq = 0.0f;

    // Loop over the batch dimension. The input is in row-major order.
    for (int i = tid; i < batch_size; i += blockDim.x) {
        int index = i * out_features + f;
        float val = static_cast<float>(x_linear[index]);
        local_sum += val;
        local_sumsq += val * val;
    }
    s_sum[tid] = local_sum;
    s_sumsq[tid] = local_sumsq;
    __syncthreads();

    // Perform reduction in shared memory to compute total sum and sumsq for feature f
    for (int s = blockDim.x / 2; s > 0; s >>= 1) {
        if (tid < s) {
            s_sum[tid] += s_sum[tid + s];
            s_sumsq[tid] += s_sumsq[tid + s];
        }
        __syncthreads();
    }

    // Compute mean and variance for feature f
    float mean = s_sum[0] / batch_size;
    float var = s_sumsq[0] / batch_size - mean * mean;

    // Update running statistics. Since one block exclusively handles feature f, no atomics are needed.
    if (tid == 0) {
        bn_running_mean[f] = bn_running_mean[f] * (1 - bn_momentum) + mean * bn_momentum;
        bn_running_var[f] = bn_running_var[f] * (1 - bn_momentum) + var * bn_momentum;
    }
    __syncthreads();

    // Second pass: apply batch normalization and Swish activation to each element in this feature column
    for (int i = tid; i < batch_size; i += blockDim.x) {
        int index = i * out_features + f;
        float val = static_cast<float>(x_linear[index]);
        float x_hat = (val - mean) / sqrtf(var + bn_eps);
        float x_bn = x_hat * bn_weight[f] + bn_bias[f];
        float x_add = x_bn + add_bias[0];
        float x_div = x_add / divide_value;
        float x_swish = x_div / (1.0f + expf(-x_div));
        output[index] = static_cast<scalar_t>(x_swish);
    }
}

// Main function: performs linear transformation then calls the fused kernel
torch::Tensor module_fn_cuda(
    torch::Tensor x,
    float bn_eps,
    float bn_momentum,
    float divide_value,
    torch::Tensor weight,
    torch::Tensor bias,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_running_mean,
    torch::Tensor bn_running_var,
    torch::Tensor add_bias) {

    const auto batch_size = x.size(0);
    const auto in_features = x.size(1);
    const auto out_features = weight.size(0);

    // Ensure tensors are contiguous
    x = x.contiguous();
    weight = weight.contiguous();
    bias = bias.contiguous();

    // Perform linear transformation: x_linear = x @ weight.T + bias
    auto x_linear = torch::addmm(bias, x, weight.t());

    // Prepare output tensor
    auto output = torch::empty_like(x_linear);

    // Launch configuration: one block per feature; each block uses a fixed number of threads
    const int threads = 256;
    const int blocks = out_features;  // one block per feature column
    const size_t shared_mem_size = 2 * threads * sizeof(float);

    AT_DISPATCH_FLOATING_TYPES(x_linear.scalar_type(), "module_fn_cuda", ([&] {
        fused_bn_swish_kernel<scalar_t><<<blocks, threads, shared_mem_size>>>(
            x_linear.data_ptr<scalar_t>(),
            output.data_ptr<scalar_t>(),
            bn_weight.data_ptr<scalar_t>(),
            bn_bias.data_ptr<scalar_t>(),
            bn_running_mean.data_ptr<scalar_t>(),
            bn_running_var.data_ptr<scalar_t>(),
            add_bias.data_ptr<scalar_t>(),
            bn_eps,
            bn_momentum,
            divide_value,
            batch_size,
            out_features);
    }));

    return output;
}

// PyBind11 module definition
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn_cuda, "Fused BN and Swish forward (CUDA)");
}
